#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import sys
import os
import multiprocessing as mp

import numpy as np
import math

#ros 
import rospy
#ros messages
from geometry_msgs.msg import Twist
from nav_msgs.msg import Odometry, Path

def q2r(q):
    """
    method to convert quaternion into rotation matrix
    @input: q .. quaternion (qx,qy,qz,qw) as np.array
    @output: R .. rotation matrix as list of lists
    """
    q = q[[3, 0, 1, 2]]
    R = [[q[0] ** 2 + q[1] ** 2 - q[2] ** 2 - q[3] ** 2, 2 * (q[1] * q[2] - q[0] * q[3]),
          2 * (q[1] * q[3] + q[0] * q[2])],
         [2 * (q[1] * q[2] + q[0] * q[3]), q[0] ** 2 - q[1] ** 2 + q[2] ** 2 - q[3] ** 2,
          2 * (q[2] * q[3] - q[0] * q[1])],
         [2 * (q[1] * q[3] - q[0] * q[2]), 2 * (q[2] * q[3] + q[0] * q[1]),
          q[0] ** 2 - q[1] ** 2 - q[2] ** 2 + q[3] ** 2]]
    return R

def r2e(r):
    """
    method to convert rotation matrix into euler angles
    @input: R .. rotation matrix as np array
    @output: e .. euler angles (roll,pitch,yaw) as list
    """
    x = math.atan2(r[2,1],r[2,2])
    y = math.atan2(-r[2,0],math.sqrt(r[2,1]*r[2,1]+r[2,2]*r[2,2]))
    z = math.atan2(r[1,0],r[0,0])
    e = [x, y, z]
    return e

def q2e(q):
    """
    method to convert quaternion into euler angles
    @input: q .. quaternion (qx,qy,qz,qw) as np.array
    @output: e .. euler angles (roll,pitch,yaw) as list
    """
    R = q2r(q)
    e = r2e(np.array(R))
    return e

class PathFollower:

    def __init__(self, path_topic_name, odometry_topic_name, velocity_topic_name, distance_thr, velocity_scale):

        self.active_path = None

        self.velocity_publisher = rospy.Publisher(velocity_topic_name, Twist, queue_size=1)
        self.path_subscriber = rospy.Subscriber(path_topic_name, Path, callback=self.path_callback, queue_size=1)
        self.odometry_subscriber = rospy.Subscriber(odometry_topic_name, Odometry, callback=self.odometry_callback, queue_size=1)

        self.distance_thr = distance_thr
        self.velocity_scale = velocity_scale

        self.path_lock = mp.Lock()

    def path_callback(self, msg):

        rospy.loginfo("%s : path callback" % rospy.get_name())
        self.path_lock.acquire()
        self.active_path = msg
        self.path_lock.release()

    def odometry_callback(self, msg):

        rospy.loginfo("%s : odometry callback" % rospy.get_name())
        cmd = Twist()

        # get current robot pose
        pos = [msg.pose.pose.position.x,msg.pose.pose.position.y]
        q = [ msg.pose.pose.orientation.x, msg.pose.pose.orientation.y, msg.pose.pose.orientation.z, msg.pose.pose.orientation.w ] 
        e = q2e(np.array(q))
        yaw_curr = e[2]

        goal = None

        # remove waypoints that are close to the robot
        self.path_lock.acquire()
        while self.active_path is not None and len(self.active_path.poses):

            goal = self.active_path.poses[0]
            coord = [goal.pose.position.x,goal.pose.position.y]

            # get manhattan distance components
            dx = coord[0] - pos[0] 
            dy = coord[1] - pos[1]

            # stop removing waypoints if distance to waypoint is too high
            if dx**2 + dy**2 >= self.distance_thr**2:
                break

            # remove waypoint from path
            self.active_path.poses = self.active_path.poses[1:]
        self.path_lock.release()

        if goal is None:

            # stop
            cmd.linear.x = 0
            cmd.angular.z = 0

        else:

            # get orientation to the target
            phi_tar = math.atan2(dy,dx);
            if phi_tar < 0:
                phi_tar += 2*math.pi

            # difference of heading to the target
            dphi = phi_tar - yaw_curr

            # make it from -pi to pi
            dphi = (dphi + math.pi) % (2*math.pi) - math.pi
        
            # put together the command
            if dphi > math.pi/16:
                cmd.linear.x = 0
                cmd.angular.z = 1. * self.velocity_scale
            elif dphi < -math.pi/16:
                cmd.linear.x = 0
                cmd.angular.z = -1. * self.velocity_scale
            else:
                cmd.linear.x = 1. * self.velocity_scale
                cmd.angular.z = 0.

        # publish the command 
        self.velocity_publisher.publish(cmd)


if __name__ == '__main__':
    
    # init ros node
    rospy.init_node('uir_path_follower', log_level=rospy.DEBUG)
    rospy.loginfo("Starting UIR path following node")
   
    # load parameters
    path_topic_name = rospy.get_param("~path_topic_name","/path")
    odometry_topic_name = rospy.get_param("~odometry_topic_name","/odometry")
    velocity_topic_name = rospy.get_param("~velocity_topic_name","/cmd_vel")
    distance_threshold = rospy.get_param("~distance_threshold",0.3) # in m
    velocity_scale = rospy.get_param("~velocity_scale",1.0) 

    # init path follower
    follower = PathFollower(path_topic_name, odometry_topic_name, velocity_topic_name, distance_thr = distance_threshold, velocity_scale = velocity_scale)

    # report loop
    rate = rospy.Rate(1)
    while not rospy.is_shutdown():
        rate.sleep()
        rospy.loginfo("%s : I am alive" % rospy.get_name())
    

