#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import sys
import os
import time

import numpy as np
import math

#ros 
import rospy
#ros messages
from geometry_msgs.msg import Twist
from nav_msgs.msg import Odometry
from sensor_msgs.msg import LaserScan
from rosgraph_msgs.msg import Clock
import tf

from uir_robot_controller import RobotRos as robot
import threading as thread
import multiprocessing as mp

class TimeHolder:

    def __init__(self):
        self.sec = 0
        self.nsec = 0
        self.lock = mp.Lock()

    def set_time(self,sec,nsec):
        self.lock.acquire()
        self.sec = sec
        self.nsec = nsec
        self.lock.release()

    def get_time(self):
        self.lock.acquire()
        sec = self.sec
        nsec = self.nsec
        self.lock.release()
        return sec,nsec

def tick_callback(msg, time_handle):
    time_handle.set_time(msg.clock.secs,msg.clock.nsecs)


def velocity_callback(msg, robot_handle):
    
    linear_x = msg.linear.x
    linear_y = msg.linear.y
    angular_z = msg.angular.z
    
    right = (linear_x + angular_z)/2
    left = linear_x - right
    
    right = min(right, 1)
    right = max(right, -1)
    left = min(left, 1)
    left = max(left, -1)

    #robot_handle.steering_lock.acquire()
    robot_handle.v_left = left
    robot_handle.v_right = right
    #robot_handle.steering_lock.release()
    rospy.loginfo("%s : steering - left: %s, right: %s" % (rospy.get_name(), str(robot_handle.v_left), str(robot_handle.v_right)) )


if __name__ == '__main__':

    # init ros node
    rospy.init_node('uir_robot_controller', log_level=rospy.DEBUG)
    rospy.loginfo("Starting Robot Control node")
    
    # load parameters
    robot_tf_name = rospy.get_param("~robot_tf_name","robot")
    odometry_topic_name = rospy.get_param("~odometry_topic_name","/odometry")
    lidar_topic_name = rospy.get_param("~lidar_topic_name","/laser_scan")
    velocity_topic_name = rospy.get_param("~velocity_topic_name","/cmd_vel")

    # robot startup
    time_holder = TimeHolder()
    rob = robot.Robot()
    rob.turn_on()
    try:
        rob.stop = False
        thread1 = thread.Thread(target=rob.locomotion)
    except:
        print("Error: unable to start locomotion thread")
        sys.exit(1)
    thread1.start()
    
    # subscribers
    vel_subscriber = rospy.Subscriber(velocity_topic_name,Twist,callback=velocity_callback,callback_args=rob,queue_size=1)
    clock_subscriber = rospy.Subscriber("/clock", Clock, callback=tick_callback, callback_args=time_holder, queue_size=1)

    # publishers
    laser_publisher = rospy.Publisher(lidar_topic_name, LaserScan, queue_size=1)
    odom_publisher = rospy.Publisher(odometry_topic_name, Odometry, queue_size=1)
    tf_broadcaster = tf.TransformBroadcaster()

    rate = rospy.Rate(10)
    while not rospy.is_shutdown():

        rospy.loginfo("%s : I am alive" % rospy.get_name());

        sec,nsec = time_holder.get_time()

        odometry_msg = rob.robot.get_robot_odometry()
        laser_scan_msg = rob.robot.get_laser_scan()

        if laser_scan_msg is not None and odometry_msg is not None:
            odometry_msg.header.stamp.secs = sec
            odometry_msg.header.stamp.nsecs = nsec
            laser_scan_msg.header.stamp.secs = sec
            laser_scan_msg.header.stamp.nsecs = nsec
            laser_scan_msg.header.frame_id = robot_tf_name
            laser_publisher.publish(laser_scan_msg)
            odom_publisher.publish(odometry_msg)
            t = (odometry_msg.pose.pose.position.x, odometry_msg.pose.pose.position.y, odometry_msg.pose.pose.position.z) 
            q = (odometry_msg.pose.pose.orientation.x, odometry_msg.pose.pose.orientation.y, odometry_msg.pose.pose.orientation.z, odometry_msg.pose.pose.orientation.w) 
            tf_broadcaster.sendTransform(t, q, odometry_msg.header.stamp, robot_tf_name, 'map')
                    
        rate.sleep()
        
    rob.stop = True
    thread1.join()

